/*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
 * TreeDiam.C -
 *     Implementation of algorithms for computing the diameter 
 * using tree heirarchies.
 * 
 * Copyright 2000 Sariel Har-Peled (ssaarriieell@cs.uiuc.edu)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * Code is based on the paper:
 *   A Practical Approach for Computing the Diameter of a Point-Set.
 *   Sariel Har-Peled (http://www.uiuc.edu/~sariel)
\*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*/

#include  <iostream>
#include  <assert.h>
#include  <stdlib.h>
#include  <stdio.h>
#include  <time.h>
#include  <math.h>

#include  <queue>


//#include  "dmalloc.h"

#include  "globjwin.h"

#include  "_generic.h"
#include  "sarray.h"
#include  "real.h"
#include  "point.h"
#include  "bbox.h"
#include  "ftree.h"
#include  "ftree2d.h"
#include  "ftreeptr.h"
#include  "TreeDiam.h"
#include  "mat3d.h"
#include  "misc.h"
//#include  "terrain.h"


#define  HEAP_LIMIT  100000
#define  HEAP_DELTA  100000

//class  heap_pairs;
//class  FSPairArray;
struct lt_fspair
{
  bool operator()(const FSPair  & a , const FSPair  & b ) const
  {
      bool  a_iis, b_iis;

      if  ( ( ! a.isIdenticalSides() )  
            &&  ( ! b.isIdenticalSides() ) )
          return   a.maxDiam() < b.maxDiam();
      
      a_iis = a.isIdenticalSides();
      b_iis = b.isIdenticalSides();
      if  ( a_iis  &&  b_iis )
          return   a.maxDiam() < b.maxDiam();
      if  ( a_iis  &&  (! b_iis ) )
          return  false;
      else
          return  true;
      //return   a.minAprxDiam() < b.minAprxDiam();
  }
};
struct lt_fspair_wspd
{
  bool operator()(const FSPair  & a , const FSPair  & b ) const
  {
      return   a.maxError() < b.maxError();
      //return   a.minAprxDiam() < b.minAprxDiam();
  }
};
struct lt_fsppair
{
    bool operator()(const FSPPair  & a , const FSPPair  & b ) const
    {
        return   a.maxDiam() < b.maxDiam();
        //return   a.minAprxDiam() < b.minAprxDiam();
    }
};
struct lt_fspair_2d
{
  bool operator()(const FS2dPair  & a , const FS2dPair  & b ) const
  {
      return   a.maxDiam() < b.maxDiam();
      //return   a.minAprxDiam() < b.minAprxDiam();
  }
};

/*--- Constants ---*/
#define   THRESHOLD_LOCAL   400
#define  THRESHOLD   10

class  heap_pairs : public std::priority_queue<FSPair, 
                    std::vector<FSPair>, lt_fspair>
{
};

//   heap_pairs;
class  FSPairArray : public  Array<FSPair> {};

/*--- Start of Code ---*/

real   brute_diameter_range( const ArrPoint3d  & arr,
                             int  start,
                             int  end,
                             PointPair  & diam )
{
    for  ( int  ind = start; ind < end; ind++ )
        for  ( int  jnd = ind + 1; jnd <= end; jnd++ ) 
            diam.update_diam( arr[ ind ], arr[ jnd ] );

    return  diam.distance;
}


#define   THRESHOLD_LOCAL   400

real   brute_diameter_local_bi( const ArrPoint3d  & arr,
                                int  a_start,
                                int  a_end,
                                int  b_start,
                                int  b_end,
                                PointPair  & diam )
{
    int   a_mid, b_mid;

    if  ( ( a_end - a_start ) < THRESHOLD_LOCAL ) {
        for  ( int  ind = a_start; ind <= a_end; ind++ )
            for  ( int  jnd = b_start; jnd <= b_end; jnd++ ) 
                diam.update_diam( arr[ ind ], arr[ jnd ] );

        return  diam.distance;
    }

    a_mid = (a_start + a_end) / 2;
    b_mid = (b_start + b_end) / 2;
    brute_diameter_local_bi( arr, 
                             a_start, a_mid, 
                             b_start, b_mid, diam );
    brute_diameter_local_bi( arr,
                             a_start, a_mid, 
                             b_mid + 1, b_end, diam );
    brute_diameter_local_bi( arr,
                             a_mid + 1, a_end, 
                             b_start, b_mid, diam );
    brute_diameter_local_bi( arr,
                             a_mid + 1, a_end,
                             b_mid + 1, b_end, diam );

    return  diam.distance;
}



real   brute_diameter_local( const ArrPoint3d  & arr,
                             int  start,
                             int  end,
                             PointPair  & out )
{
    int  mid;

    if  ( ( end - start ) < THRESHOLD_LOCAL )
        return   brute_diameter_range( arr, start, end, out );
        
    mid = ( start + end ) / 2;

    brute_diameter_local_bi( arr, start, mid, mid + 1, end, out );
    brute_diameter_local( arr, start, mid, out );
    brute_diameter_local( arr, mid + 1, end, out );

    return  out.distance;
}


real   brute_diameter_local( const ArrPoint3d  & arr,
                             PointPair  & out )
{
    out.p = out.q = arr[ 0 ];
    out.distance = 0;

    return  brute_diameter_local( arr, 0, arr.size() - 1, out );
}



void   fill_array_cube( ArrPoint3d  & arr, int  num )
{
    arr.init( num + 10, 43897543 );

    for  ( int  ind = 0; ind < num; ind++ ) {
        Point3d  pnt;
        
        pnt.fillRandom();
        arr.pushx( pnt );
    }

}


void   fill_array_box( ArrPoint3d  & arr, int  num,
                       double  a, double  b, double  c )
{
    arr.init( num + 10, 43897543 );

    for  ( int  ind = 0; ind < num; ind++ ) {
        Point3d  pnt;
        
        pnt.fillRandom();
        Point3d  pntx( pnt[ 0 ] * a, pnt[ 1 ] * b, pnt[ 2 ] * c );

        arr.pushx( pntx );
    }

}


void   fill_array_sphere( ArrPoint3d  & arr, int  num )
{
    arr.init( num + 10, 43897543 );

    for  ( int  ind = 0; ind < num; ind++ ) {
        double  alpha, beta;
        Point3d  pnt;
        
        alpha = realRand() * 2.0 * 3.1415926535;
        beta = realRand() * 2.0 * 3.1415926535;

        pnt.setCoord( 0, cos( alpha ) * sin( beta ) );
        pnt.setCoord( 1, cos( alpha ) * cos( beta ) );
        pnt.setCoord( 2, sin( alpha ) );
        
        //pnt.fillRandom();
        arr.pushx( pnt );
    }

}


void   fill_array_sphere_caps( ArrPoint3d  & arr, int  num,
                            double  eps )
{
    arr.init( num + 10, 43897543 );

    for  ( int  ind = 0; ind < num; ind++ ) {
        double  alpha, beta;
        Point3d  pnt;
        
        alpha = (1.0 - eps * realRand()) * 3.1415926535 / 2.0;
        if  ( realRand() >= 0.5 ) 
            alpha = -alpha;

        beta = realRand() * 2.0 * 3.1415926535;

        pnt.setCoord( 0, cos( alpha ) * sin( beta ) );
        pnt.setCoord( 1, cos( alpha ) * cos( beta ) );
        pnt.setCoord( 2, sin( alpha ) );
        
        //pnt.fillRandom();
        arr.pushx( pnt );
    }

}


void   fill_array_two_arcs( ArrPoint3d  & arr, int  num )
{
    double  alpha_limit, alpha_increment, alpha;
    Point3d  pnt;

    arr.init( num + 10, 43897543 );

    num  = num / 2;

    alpha_limit = 1.0 /((double)num*num);
    alpha_increment = 2.0 * alpha_limit  / (double)num;
    
    Transform3d  trans_x, trans_z;
    trans_x.init_rot_around_x( 3.14150265358979323 / 4.0 );
    trans_z.init_rot_around_z( 3.14150265358979323 / 4.0 );

    alpha = -alpha_limit;
    for  ( int  ind = 0; ind < num; ind++ ) {
        pnt = Point3d( -cos( alpha ), 0, sin( alpha ) );
        trans_x.transform( pnt );
        trans_z.transform( pnt );
        arr.pushx( pnt );

        pnt = Point3d( cos( alpha ), sin( alpha ), 0 );
        trans_x.transform( pnt );
        trans_z.transform( pnt );
        arr.pushx( pnt );

        alpha += alpha_increment;
    }

}

void   fill_array_mutated_ellipsoid( ArrPoint3d  & arr, int  num,
                                     double  a, double  b, double  c )
{
    Point3d  new_pnt, pnt;

    arr.init( num + 10, 43897543 );

    Point3d  vec1( 1, 1, 1 );
    Point3d  vec2( 1, -1, 0 );
    Point3d  vec3;

    vec3 = cross_prod( vec1, vec2 );

    vec1.normalize();
    vec2.normalize();
    vec3.normalize();

    int  limit;

    limit = num / 11;
    for  ( int  ind = 0; ind < limit; ind++ ) {
        double  alpha, beta;
        Point3d  pnt;
        
        alpha = realRand() * 2.0 * 3.1415926535;
        beta = realRand() * 2.0 * 3.1415926535;

        pnt.setCoord( 0, a * cos( alpha ) * sin( beta ) );
        pnt.setCoord( 1, b * cos( alpha ) * cos( beta ) );
        pnt.setCoord( 2, c * sin( alpha ) );
        
        new_pnt = changeBase( vec1, vec2, vec3, pnt );
        
        //pnt.fillRandom();
        arr.pushx( new_pnt );
    }

    limit = 5 * num / 11;
    for  ( int  ind = 0; ind < limit; ind++ ) {
        pnt.setCoord( 0, 0 );
        pnt.setCoord( 1, 0 );
        pnt.setCoord( 2, 0.999 * c );

        new_pnt = changeBase( vec1, vec2, vec3, pnt );        
        arr.pushx( new_pnt );
    }
    for  ( int  ind = 0; ind < limit; ind++ ) {
        pnt.setCoord( 0, 0 );
        pnt.setCoord( 1, 0.999 * b );
        pnt.setCoord( 2, 0 );

        new_pnt = changeBase( vec1, vec2, vec3, pnt );        
        arr.pushx( new_pnt );
    }
}


static void   addPoint( ArrPoint3d  & arr, double  x, double  y, double  z )
{
    Point3d  pnt( x, y, z );
    Point3d  pnt1;

    pnt1.fillRandom();
    pnt.add_scale( pnt1, 0.00001 );
    arr.pushx( pnt );
}


// Cheats the regular bbox/pca algorithm. so they realize 1.72
// approximation in 3d
void   fill_array_strange_cube( ArrPoint3d  & arr, int  num )
{
    int limit;
    //Point3d  pnt, pnt1;
        
    arr.init( num + 10, 43897543 );

    addPoint( arr, -1, -1, -1 );
    addPoint( arr, -1, -1,  1 );
    addPoint( arr, -1,  1, -1 );
    addPoint( arr, -1,  1,  1 );
    addPoint( arr,  1, -1, -1 );
    addPoint( arr,  1, -1,  1 );
    addPoint( arr,  1,  1, -1 );
    addPoint( arr,  1,  1,  1 );

    addPoint( arr,  0,  0,  1.01000001 );
    addPoint( arr,  0,  0, -1.0100001 );
    addPoint( arr,  0,  1.01, 0 );
    addPoint( arr,  0, -1.01, 0 );
    addPoint( arr,  1.01, 0, 0 );
    addPoint( arr, -1.01, 0, 0 );

    limit = num / 12;
    for  ( int  ind = 0; ind < limit; ind++ ) {
        addPoint( arr,  0.001,  0,  1.011 );
        addPoint( arr,  0,  0.0001, -1.011 );
        addPoint( arr,  0.001,  0,  1.011 );
        addPoint( arr,  0,  0.0001, -1.011 );
        addPoint( arr,  0.001,  0,  1.011 );
        addPoint( arr,  0,  0.0001, -1.011 );

        addPoint( arr,  0,  1.011, 0 );
        addPoint( arr,  0, -1.011, 0 );
        addPoint( arr,  0,  1.011, 0 );
        addPoint( arr,  0, -1.011, 0 );

        addPoint( arr,   1.011, 0, 0 );
        addPoint( arr,  -1.011, 0, 0 );
    }
}



void   fill_array_strange_cube_diag( ArrPoint3d  & arr, int  num )
{
    int limit;
    //Point3d  pnt, pnt1;
        
    arr.init( num + 10, 43897543 );

    addPoint( arr, -1, -1, -1 );
    addPoint( arr, -1, -1,  1 );
    addPoint( arr, -1,  1, -1 );
    addPoint( arr, -1,  1,  1 );
    addPoint( arr,  1, -1, -1 );
    addPoint( arr,  1, -1,  1 );
    addPoint( arr,  1,  1, -1 );
    addPoint( arr,  1,  1,  1 );

    addPoint( arr,  0,  0,  1.1000001 );
    addPoint( arr,  0,  0, -1.100001 );
    addPoint( arr,  0,  1.1, 0 );
    addPoint( arr,  0, -1.1, 0 );
    addPoint( arr,  1.1, 0, 0 );
    addPoint( arr, -1.1, 0, 0 );

    limit = num / 2;
    for  ( int  ind = 0; ind < limit; ind++ ) {
        addPoint( arr, -1, -1, -1 );
        addPoint( arr,  1,  1,  1 );
    }
}


void  computeExtremePairDir( const ArrPoint3d  & arr,
                             const Point3d  & dir,
                             PointPair  & pp
                             )
{
    real  min_dot, max_dot, prod;
    int  ind_min_dot, ind_max_dot;

    pp.init( arr[ 0 ] );
    min_dot = max_dot = dir.dot_prod( arr[ 0 ] );
    ind_min_dot = ind_max_dot = 0;

    for  ( int  ind = 1; ind < arr.size(); ind++ ) {
        prod = dir.dot_prod( arr[ ind ] );
        if  ( prod < min_dot ) {
            min_dot = prod;
            ind_min_dot = ind;
        }
        if  ( prod > max_dot ) {
            max_dot = prod;
            ind_max_dot = ind;
        }
    }

    pp.init( arr[ ind_min_dot ], arr[ ind_max_dot ] );
}


void  computeExtremePairDirs( const ArrPoint3d  & arr,
                              const Point3d  & dir_0,
                              const Point3d  & dir_1,
                              const Point3d  & dir_2,
                              PointPair  & pp_0,
                              PointPair  & pp_1,
                              PointPair  & pp_2
                             )
{
    real  min_dot_0, max_dot_0, prod_0;
    int  ind_min_dot_0, ind_max_dot_0;
    real  min_dot_1, max_dot_1, prod_1;
    int  ind_min_dot_1, ind_max_dot_1;
    real  min_dot_2, max_dot_2, prod_2;
    int  ind_min_dot_2, ind_max_dot_2;

    pp_0.init( arr[ 0 ] );
    pp_1.init( arr[ 0 ] );
    pp_2.init( arr[ 0 ] );
    min_dot_0 = max_dot_0 = dir_0.dot_prod( arr[ 0 ] );
    ind_min_dot_0 = ind_max_dot_0 = 0;
 
    min_dot_1 = max_dot_1 = dir_1.dot_prod( arr[ 0 ] );
    ind_min_dot_1 = ind_max_dot_1 = 0;

    min_dot_2 = max_dot_2 = dir_2.dot_prod( arr[ 0 ] );
    ind_min_dot_2 = ind_max_dot_2 = 0;

    for  ( int  ind = 1; ind < arr.size(); ind++ ) {
        const Point3d  & pnt( arr[ ind ] );

        prod_0 = dir_0.dot_prod( pnt );
        if  ( prod_0 < min_dot_0 ) {
            min_dot_0 = prod_0;
            ind_min_dot_0 = ind;
        }
        if  ( prod_0 > max_dot_0 ) {
            max_dot_0 = prod_0;
            ind_max_dot_0 = ind;
        }

        prod_1 = dir_1.dot_prod( pnt );
        if  ( prod_1 < min_dot_1 ) {
            min_dot_1 = prod_1;
            ind_min_dot_1 = ind;
        }
        if  ( prod_1 > max_dot_1 ) {
            max_dot_1 = prod_1;
            ind_max_dot_1 = ind;
        }

        prod_2 = dir_2.dot_prod( pnt );
        if  ( prod_2 < min_dot_2 ) {
            min_dot_2 = prod_2;
            ind_min_dot_2 = ind;
        }
        if  ( prod_2 > max_dot_2 ) {
            max_dot_2 = prod_2;
            ind_max_dot_2 = ind;
        }
    }

    pp_0.init( arr[ ind_min_dot_0 ], arr[ ind_max_dot_0 ] );
    pp_1.init( arr[ ind_min_dot_1 ], arr[ ind_max_dot_1 ] );
    pp_2.init( arr[ ind_min_dot_2 ], arr[ ind_max_dot_2 ] );
}


void  computeExtremePairDir( const ArrPoint3d  & arr,
                             const Point3d  & dir,
                             PointPair  & pp,
                             int  start,
                             int  end
                             )
{
    real  min_dot, max_dot, prod;
    const Point3d  *p_pnt_min, *p_pnt_max;

    p_pnt_min = &pp.p;
    p_pnt_max = &pp.q;

    if  ( dir.dot_prod( *p_pnt_min ) > dir.dot_prod( *p_pnt_max ) )
        exchange( p_pnt_min, p_pnt_max );

    min_dot = dir.dot_prod( *p_pnt_min );
    max_dot = dir.dot_prod( *p_pnt_max );
 
    for  ( int  ind = start; ind < arr.size(); ind++ ) {
        prod = dir.dot_prod( arr[ ind ] );
        if  ( prod < min_dot ) {
            min_dot = prod;
            p_pnt_min = &( arr[ ind ] );
        }
        if  ( prod > max_dot ) {
            max_dot = prod;
            p_pnt_max = &( arr[ ind ] );
        }
    }
    Point3d  p, q;

    p = *p_pnt_max;
    q = *p_pnt_min;
    pp.init( p, q );
}



void  computeExtremePair( const ArrPoint3d  & arr,
                          int  dim, PointPair  & pp )
{
    pp.init( arr[ 0 ] );

    for  ( int  ind = 1; ind < arr.size(); ind++ ) {
        const Point3d  & pnt( arr[ ind ] );

        if  ( pnt[ dim ] < pp.p[ dim ] )
            pp.p = pnt;
        if  ( pnt[ dim ] > pp.q[ dim ] )
            pp.q = pnt;
    }

    pp.distance = pnt_distance( pp.p, pp.q );
}


void  computeExtremePairAllDim( const ArrPoint3d  & arr,
                                PointPair  & pp0,
                                PointPair  & pp1,
                                PointPair  & pp2
                                )
{
    pp0.init( arr[ 0 ] );
    pp1.init( arr[ 0 ] );
    pp2.init( arr[ 0 ] );

    for  ( int  ind = 1; ind < arr.size(); ind++ ) {
        const Point3d  & pnt( arr[ ind ] );
        
        if  ( pnt[ 0 ] < pp0.p[ 0 ] )
            pp0.p = pnt;
        if  ( pnt[ 0 ] > pp0.q[ 0 ] )
            pp0.q = pnt;

        if  ( pnt[ 1 ] < pp1.p[ 1 ] )
            pp1.p = pnt;
        if  ( pnt[ 1 ] > pp1.q[ 1 ] )
            pp1.q = pnt;

        if  ( pnt[ 2 ] < pp2.p[ 2 ] )
            pp2.p = pnt;
        if  ( pnt[ 2 ] > pp2.q[ 2 ] )
            pp2.q = pnt;
    }

    pp0.distance = pnt_distance( pp0.p, pp0.q );
    pp1.distance = pnt_distance( pp1.p, pp1.q );
    pp2.distance = pnt_distance( pp2.p, pp2.q );
}


void  computeExtremePair( CPoint3dPtr  arr,
                          int  size,
                          int  dim, PointPair  & pp )
{
    //printf( "arr: %p size: %d\n", arr, size );

    pp.init( arr[ 0 ] );

    for  ( int  ind = 1; ind < size; ind++ ) {
        const Point3d  & pnt( arr[ ind ] );

        if  ( pnt[ dim ] < pp.p[ dim ] )
            pp.p = pnt;
        if  ( pnt[ dim ] > pp.q[ dim ] )
            pp.q = pnt;
    }

    pp.distance = pnt_distance( pp.p, pp.q );
}


void  TreeDiamAlg::compute_by_level( double  eps ) 
{
    FSPairArray  queue_a, queue_b;
    FSPairArray  * p_queue_in, * p_queue_out;
    int  heap_limit, heap_delta;
    
    queue_a.init( 100, 43432 );
    queue_b.init( 100, 43431 );
    
    FSTreeNode  * root = tree.getRoot();

    computeExtremePair( tree.getPoints(), root->getBBox().getLongestDim(),
                        pair_diam );

    FSPair  root_pair;
    
    root_pair.init( root, root );
    
    queue_a.pushx( root_pair );
    
    heap_limit = HEAP_LIMIT;
    heap_delta = HEAP_DELTA;
    p_queue_in = &queue_a;
    p_queue_out = &queue_b;
    for  ( int  ind = 0; ind < 1000; ind++ ) {
        //printf( "ind: %d\n", ind );
        if   ( p_queue_in->isEmpty() ) {
            //printf( "breaking!\n" );
            break;
        }
        p_queue_out->reset();
        if  ( p_queue_in->size() > heap_limit ) {
                threshold_brute *= 4;
                printf( "threshold_brute: %d\n", threshold_brute );
                heap_limit += heap_delta;
        }

        split_level( *p_queue_in, *p_queue_out, eps );
        
        exchange( p_queue_in, p_queue_out );
    }
    
}


void  TreeDiamAlg::compute_by_heap( double  eps ) 
{
    heap_pairs  heap;
    int  heap_limit, heap_delta;
    
    heap_limit = HEAP_LIMIT;
    heap_delta = HEAP_DELTA;

    printf( "lll\n" );
    FSTreeNode  * root = tree.getRoot();
    
    computeExtremePair( tree.getPoints(), root->getBBox().getLongestDim(),
                        pair_diam );
    //diam = root->getBBox().getLongestEdge();
    
    FSPair  root_pair;
    
    root_pair.init( root, root );
    
    printf( "root diam: %g\n", root_pair.maxDiam() );
    heap.push( root_pair );
    //queue_a.pushx( root_pair );
    
    int  count =0;
    while  ( heap.size() > 0 ) {
        //printf( "%d: curr_diam: %g\n", count++, pair_diam.distance );
        FSPair  pair = heap.top();
        split_pair( pair, heap, eps );
    
        if  ( ( count & 0x3ff ) == 0 ) {
            if  ( ((int)heap.size()) > heap_limit ) {
                threshold_brute *= 2;
                printf( "threshold_brute: %d\n", threshold_brute );
                heap_limit += heap_delta;
            }
            //printf( "size: %d  [%d]\n", (int)heap.size(),
            //        (int)sizeof( FSPair ) );
            //mem_report();
            //fflush( stdout );
        }

        count++;
        heap.pop();
    }
    printf( "\tOverall count:%d\n", count );
}


void  TreeDiamAlg::compute_by_heap_wspd( double  eps ) 
{
    heap_pairs  heap;
    int  count_self_pairs;
    int  heap_limit, heap_delta;
    
    heap_limit = HEAP_LIMIT;
    heap_delta = HEAP_DELTA;

    FSTreeNode  * root = tree.getRoot();
    
    computeExtremePair( tree.getPoints(), root->getBBox().getLongestDim(),
                        pair_diam );
    //diam = root->getBBox().getLongestEdge();
    
    FSPair  root_pair;
    
    root_pair.init( root, root );
    
    heap.push( root_pair );
    //queue_a.pushx( root_pair );
    
    //t  count =0;
    int  count = 0;

    count_self_pairs = 1;
    while  ( count_self_pairs > 0 ) {
        //intf( "%d: curr_diam: %g\n", count++, pair_diam.distance );
        FSPair  pair = heap.top();

        if  ( ( count & 0x3ff ) == 0 ) {
            if  ( ((int)heap.size()) > heap_limit ) {
                threshold_brute *= 2;
                printf( "threshold_brute: %d\n", threshold_brute );
                heap_limit += heap_delta;
            } 
       }

        if  ( pair.isIdenticalSides() )
            --count_self_pairs;
        split_pair_wspd( pair, heap, eps, count_self_pairs );
        count++;
        
        heap.pop();
    }

    //printf( "no self pairs in heap!\ncount:%d\n", count );

    while  ( heap.size() > 0 ) {
        //intf( "%d: curr_diam: %g\n", count++, pair_diam.distance );
        FSPair  pair = heap.top();

        if  ( ( count & 0x3ff ) == 0 ) {
            if  ( ((int)heap.size()) > heap_limit ) {
                threshold_brute *= 2;
                printf( "threshold_brute: %d\n", threshold_brute );
                heap_limit += heap_delta;
            } 
       }

        ///split_pair_wspd( pair, heap, eps, count_self_pairs );
        split_pair_wspd_one_side( pair, heap, eps, count_self_pairs );
        count++;

        heap.pop();
    }
    printf( "\tOverall count:%d\n", count );
}


void  TreeDiamAlg::split_pair_dir( FSPair  & pair,
                                   heap_pairs  & heap, 
                                   double  eps,
                                   double  threshold )
{
    bool  f_is_left_splitable, f_is_right_splitable;
    
    //printf( "split_pair_dir: (%g), ?, %g, %g\n", 
    //        pair_diam.distance, eps, threshold );
    if  ( pair.maxDiam() <= ((1.0 + eps) * pair_diam.distance ))
        return;
    
    if  ( pair.isBelowThreshold( threshold_brute ) ) {
        pair.directDiameter( tree, pair_diam );
        return;
    }
    if  ( pair.getProjectionError() <= threshold ) {
        tree.findExtremeInProjection( pair, pair_diam );
        return;
    }

    f_is_left_splitable = pair.get_left()->isSplitable();
    f_is_right_splitable = pair.get_right()->isSplitable();
    assert( f_is_left_splitable  ||  f_is_right_splitable );
    if ( f_is_left_splitable )
        tree.split_node( pair.get_left() );
    if ( f_is_right_splitable )
        tree.split_node( pair.get_right() );
    
    if  ( f_is_left_splitable 
          &&  f_is_right_splitable ) {
        addPairHeap( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right()->get_right() );
        // to avoid exponential blowup
        if  ( pair.get_left() != pair.get_right() )
            addPairHeap( heap, 
                         pair.get_left()->get_right(),
                         pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right()->get_right() );
        return;
    }
    if  ( f_is_left_splitable ) {
        addPairHeap( heap,
                     pair.get_left()->get_left(),
                     pair.get_right() );
        addPairHeap( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right() );
        return;
    }
    if  ( f_is_right_splitable ) {
        addPairHeap( heap, 
                     pair.get_left(),
                     pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left(),
                     pair.get_right()->get_right() );
        return;
    }
}

void  TreeDiamAlg::compute_by_dir( double  eps ) 
{
    heap_pairs  heap;
    double  threshold;
    int  heap_limit, heap_delta;
    
    heap_limit = HEAP_LIMIT;
    heap_delta = HEAP_DELTA;

    //printf( "compute_by_dir: %.4f\n", eps );
    // if a pair have 2r/l <= threshold then the error is small
    // enough, and one can use the direction between the centers to do
    // the computations.
    threshold = sqrt( 1.0 - (1.0 - eps) * (1.0 - eps) );
    FSTreeNode  * root = tree.getRoot();
    
    computeExtremePair( tree.getPoints(), root->getBBox().getLongestDim(),
                        pair_diam );
    //diam = root->getBBox().getLongestEdge();
    
    FSPair  root_pair;
    
    root_pair.init( root, root );
    
    heap.push( root_pair );
    //queue_a.pushx( root_pair );
    
    int  count =0;
    while  ( heap.size() > 0 ) {
        //printf( "%d: curr_diam: %g\n", count++, pair_diam.distance );
        count++;
        FSPair  pair = heap.top();
        if  ( ( count & 0x3ff ) == 0 ) {
            if  ( ((int)heap.size()) > heap_limit ) {
                threshold_brute *= 2;
                printf( "threshold_brute: %d\n", threshold_brute );
                heap_limit += heap_delta;
            } 
        }

        split_pair_dir( pair, heap, eps, threshold );
        
        heap.pop();
    }
}


void  TreeDiamAlg::addPair( FSPairArray  & queue_out, 
                            FSTreeNode  * left,
                            FSTreeNode  * right )
{
    const Point3d  & p( tree.getPoint( left->index_left() ) );
    const Point3d  & q( tree.getPoint( right->index_left() ) );
    
    //printf( "old_diam; %g\n", diam );
    pair_diam.update_diam( p, q );

    //diam = max( diam, pnt_distance( p, q ) );
    //printf( "new_diam; %g\n", diam );
    
    FSPair  pair;
    
    pair.init( left, right );
    if  ( pair.maxDiam() <= pair_diam.distance )
        return;
    queue_out.pushx( pair );
}


void  TreeDiamAlg::split_level( FSPairArray  & queue_in,
                                FSPairArray  & queue_out, 
                                double  eps )
{
    bool  f_is_left_splitable, f_is_right_splitable;
    
    for  ( int  ind = 0; ind < queue_in.size(); ind++ ) {
        FSPair  & in( queue_in[ ind ] );
        if  ( in.maxDiam() <= (1.0 + eps) * pair_diam.distance )
            continue;
        
        if  ( in.isBelowThreshold( threshold_brute ) ) {
            in.directDiameter( tree, pair_diam );
            continue;
        }
        
        //printf( "\n---curr_diam: %g\n", diam );
        //in.dump();
        //printf( "-------------\n\n" );
        // Oh well, we need to split, if possible...
        f_is_left_splitable = in.get_left()->isSplitable();
        f_is_right_splitable = in.get_right()->isSplitable();
        assert( f_is_left_splitable ||  f_is_right_splitable );
        if ( f_is_left_splitable )
            tree.split_node( in.get_left() );
        if ( f_is_right_splitable )
            tree.split_node( in.get_right() );
        
        if  ( f_is_left_splitable 
              &&  f_is_right_splitable ) {
            addPair( queue_out, 
                     in.get_left()->get_left(),
                     in.get_right()->get_left() );
            addPair( queue_out, 
                     in.get_left()->get_left(),
                     in.get_right()->get_right() );
            // to avoid exponential blowup
            if  ( in.get_left() != in.get_right() )
                addPair( queue_out, 
                         in.get_left()->get_right(),
                         in.get_right()->get_left() );
            addPair( queue_out, 
                     in.get_left()->get_right(),
                     in.get_right()->get_right() );
            continue;
        }
        if  ( f_is_left_splitable ) {
            addPair( queue_out, 
                     in.get_left()->get_left(),
                     in.get_right() );
            addPair( queue_out, 
                     in.get_left()->get_right(),
                     in.get_right() );
            continue;
        }
        if  ( f_is_right_splitable ) {
            addPair( queue_out, 
                     in.get_left(),
                     in.get_right()->get_left() );
            addPair( queue_out, 
                     in.get_left(),
                     in.get_right()->get_right() );
            continue;
        }
    }
}


void  TreeDiamAlg::split_pair( FSPair  & pair,
                               heap_pairs  & heap, double  eps )
{
    bool  f_is_left_splitable, f_is_right_splitable;
    
    if  ( pair.maxDiam() <= ((1.0 + eps) * pair_diam.distance ))
        return;
    
    //printf( "threshold_brute: %d\n", threshold_brute );
    if  ( pair.isBelowThreshold( threshold_brute ) ) {
        pair.directDiameter( tree, pair_diam );
        return;
    }
    
    f_is_left_splitable = pair.get_left()->isSplitable();
    f_is_right_splitable = pair.get_right()->isSplitable();
    assert( f_is_left_splitable  ||  f_is_right_splitable );
    if ( f_is_left_splitable )
        tree.split_node( pair.get_left() );
    if ( f_is_right_splitable )
        tree.split_node( pair.get_right() );
    
    if  ( f_is_left_splitable 
          &&  f_is_right_splitable ) {
        addPairHeap( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right()->get_right() );
        // to avoid exponential blowup
        if  ( pair.get_left() != pair.get_right() )
            addPairHeap( heap, 
                         pair.get_left()->get_right(),
                         pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right()->get_right() );
        return;
    }
    if  ( f_is_left_splitable ) {
        addPairHeap( heap,
                     pair.get_left()->get_left(),
                     pair.get_right() );
        addPairHeap( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right() );
        return;
    }
    if  ( f_is_right_splitable ) {
        addPairHeap( heap, 
                     pair.get_left(),
                     pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left(),
                     pair.get_right()->get_right() );
        return;
    }
}


void  TreeDiamAlg::split_pair_wspd( FSPair  & pair,
                                    heap_pairs  & heap, 
                                    double  eps,
                                    int  & count_self_pairs )
{
    bool  f_is_left_splitable, f_is_right_splitable;
    
    if  ( pair.maxDiam() <= ((1.0 + eps) * pair_diam.distance ))
        return;
    
    if  ( pair.isBelowThreshold( threshold_brute ) ) {
        pair.directDiameter( tree, pair_diam );
        return;
    }
    
    f_is_left_splitable = pair.get_left()->isSplitable();
    f_is_right_splitable = pair.get_right()->isSplitable();
    assert( f_is_left_splitable  ||  f_is_right_splitable );
    if ( f_is_left_splitable )
        tree.split_node( pair.get_left() );
    if ( f_is_right_splitable )
        tree.split_node( pair.get_right() );
    
    if  ( f_is_left_splitable 
          &&  f_is_right_splitable ) {
        addPairHeapWSPD( heap, 
                         pair.get_left()->get_left(),
                         pair.get_right()->get_left(),
                         eps,
                         count_self_pairs );
        addPairHeapWSPD( heap, 
                         pair.get_left()->get_left(),
                         pair.get_right()->get_right(),
                         eps,
                         count_self_pairs  );
        // to avoid exponential blowup
        if  ( pair.get_left() != pair.get_right() )
            addPairHeapWSPD( heap, 
                             pair.get_left()->get_right(),
                             pair.get_right()->get_left(),
                             eps,
                             count_self_pairs  );
        addPairHeapWSPD( heap, 
                         pair.get_left()->get_right(),
                         pair.get_right()->get_right(),
                         eps,
                         count_self_pairs  );
        return;
    }
    if  ( f_is_left_splitable ) {
        addPairHeapWSPD( heap,
                         pair.get_left()->get_left(),
                         pair.get_right(),
                         eps,
                         count_self_pairs  );
        addPairHeapWSPD( heap, 
                         pair.get_left()->get_right(),
                         pair.get_right(),
                         eps,
                         count_self_pairs  );
        return;
    }
    if  ( f_is_right_splitable ) {
        addPairHeapWSPD( heap, 
                         pair.get_left(),
                         pair.get_right()->get_left(),
                         eps,
                         count_self_pairs  );
        addPairHeapWSPD( heap, 
                         pair.get_left(),
                         pair.get_right()->get_right(),
                         eps,
                         count_self_pairs  );
        return;
    }
}



// Splits a pair only one side - avoiding a huge number of pairs...?
void  TreeDiamAlg::split_pair_wspd_one_side( FSPair  & pair,
                                             heap_pairs  & heap, 
                                             double  eps,
                                             int  & count_self_pairs )
{
    bool  f_is_left_splitable;

    if  ( pair.maxDiam() <= ((1.0 + eps) * pair_diam.distance ))
        return;
    
    if  ( pair.isBelowThreshold( threshold_brute ) ) {
        pair.directDiameter( tree, pair_diam );
        return;
    }
    
    pair.setLeftBigger();

    f_is_left_splitable = pair.get_left()->isSplitable();
    assert( f_is_left_splitable );
    if ( f_is_left_splitable )
        tree.split_node( pair.get_left() );
    
    addPairHeapWSPD( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right(),
                     eps,
                     count_self_pairs );
    addPairHeapWSPD( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right(),
                     eps,
                     count_self_pairs  );
}


real   brute_diameter( const ArrPoint3d  & arr,
                       PointPair  & diam )
{
    diam.init( arr[ 0 ] );

    for  ( int  ind = 0; ind < ( arr.size() - 1 ); ind++ ) {
        Point3d  pnt( arr[ ind ] );
 
        for  ( int  jnd = ind + 1; jnd < arr.size(); jnd++ ) 
            diam.update_diam( pnt, arr[ jnd ] );
    }

    return  diam.distance;
}

void  TreeDiamAlg::addPairHeap( heap_pairs  & heap,
                             FSTreeNode  * left,
                             FSTreeNode  * right )
{
    const Point3d  & p( tree.getPoint( left->index_left() ) );
    const Point3d  & q( tree.getPoint( right->index_left() ) );


    pair_diam.update_diam( p, q );
    //printf( "old_diam; %g\n", diam );
    //diam = max( diam, pnt_distance( p, q ) );
    //printf( "new_diam; %g\n", diam );
    
    FSPair  pair;
    
    pair.init( left, right );
    //printf( "__pair max diameter: %g\n", pair.maxDiam() );
    
    if  ( pair.maxDiam() <= pair_diam.distance )
        return;
    //printf( "pair max diameter: %g\n", pair.maxDiam() );
    heap.push( pair );
}


void  TreeDiamAlg::addPairHeapWSPD( heap_pairs  & heap,
                                    FSTreeNode  * left,
                                    FSTreeNode  * right,
                                    double  eps,
                                    int  & count_self_pairs
                                    )
{
    const Point3d  & p( tree.getPoint( left->index_left() ) );
    const Point3d  & q( tree.getPoint( right->index_left() ) );

    pair_diam.update_diam( p, q );
    //printf( "old_diam; %g\n", diam );
    //diam = max( diam, pnt_distance( p, q ) );
    //printf( "new_diam; %g\n", diam );
    
    FSPair  pair;
    
    pair.init( left, right );
    if  ( pair.maxDiam() <= (1.0 + eps) * pair_diam.distance )
        return;
    //printf( "iis: %d, pair diam: %10g  estimate: %g\n", 
    //        (int)( pair.isIdenticalSides() ),
    //        pair.maxDiam(),
    //        pair_diam.distance );
    heap.push( pair );
    if  ( pair.isIdenticalSides() )
        count_self_pairs++;
}


void  pnt_array_bound( ArrPoint3d  & arr, BBox3d  & bb )
{
    bb.init();

    for  ( int  ind = 0; ind < arr.size(); ind++ ) 
        bb.bound( arr[ ind ] );
}


void  FSTree::findExtremeInProjection( FSPair  & pair, 
                                       PointPair  & max_pair_diam )
{
    PointPair  candid;
    Point3d  dir;
    
    dir = pnt_subtract( pair.get_left()->getCenter(), 
                        pair.get_right()->getCenter() );
    
    candid.init( arr[ pair.get_left()->index_left() ] );
    computeExtremePairDir( arr, dir, candid, 
                           pair.get_left()->index_left(),
                           pair.get_left()->index_right() );
    computeExtremePairDir( arr, dir, candid, 
                           pair.get_right()->index_left(),
                           pair.get_right()->index_right() );
    max_pair_diam.update_diam( candid );
}


//--------------------------------------------------------------------------
// TreeDiamAlg2d
//--------------------------------------------------------------------------

void  computeExtremePair( const ArrPoint2dPtr  & arr,
                          int  dim, PointPair2d  & pp )
{
    pp.init( *(arr[ 0 ]) );

    for  ( int  ind = 1; ind < arr.size(); ind++ ) {
        const Point2d  & pnt( *(arr[ ind ]) );

        if  ( pnt[ dim ] < pp.p[ dim ] )
            pp.p = pnt;
        if  ( pnt[ dim ] > pp.q[ dim ] )
            pp.q = pnt;
    }

    pp.distance = pnt_distance( pp.p, pp.q );
}

class  heap_pairs_2d :
          public std::priority_queue<FS2dPair, 
          std::vector<FS2dPair>, lt_fspair_2d>
{
};

void  TreeDiamAlg2d::addPairHeap( heap_pairs_2d  & heap,
                                  FS2dTreeNode  * left,
                                  FS2dTreeNode  * right )
{
    const Point2d  & p( tree.getPoint( left->index_left() ) );
    const Point2d  & q( tree.getPoint( right->index_left() ) );

    pair_diam.update_diam( p, q );
    //printf( "old_diam; %g\n", diam );
    //diam = max( diam, pnt_distance( p, q ) );
    //printf( "new_diam; %g\n", diam );
    
    FS2dPair  pair;
    
    pair.init( left, right );
    if  ( pair.maxDiam() <= pair_diam.distance )
        return;
    heap.push( pair );
}


void  TreeDiamAlg2d::split_pair( FS2dPair  & pair,
                                 heap_pairs_2d  & heap, double  eps )
{
    bool  f_is_left_splitable, f_is_right_splitable;

    //printf( "xx\n" );
    if  ( pair.maxDiam() <= ((1.0 + eps) * pair_diam.distance ))
        return;
    
    if  ( pair.isBelowThreshold( threshold_brute ) ) {
        pair.directDiameter( tree, pair_diam );
        return;
    }
    
    f_is_left_splitable = pair.get_left()->isSplitable();
    f_is_right_splitable = pair.get_right()->isSplitable();
    assert( f_is_left_splitable  ||  f_is_right_splitable );
    if ( f_is_left_splitable )
        tree.split_node( pair.get_left() );
    if ( f_is_right_splitable )
        tree.split_node( pair.get_right() );
    
    if  ( f_is_left_splitable 
          &&  f_is_right_splitable ) {
        addPairHeap( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right()->get_right() );
        // to avoid exponential blowup
        if  ( pair.get_left() != pair.get_right() )
            addPairHeap( heap, 
                         pair.get_left()->get_right(),
                         pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right()->get_right() );
        return;
    }
    if  ( f_is_left_splitable ) {
        addPairHeap( heap,
                     pair.get_left()->get_left(),
                     pair.get_right() );
        addPairHeap( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right() );
        return;
    }
    if  ( f_is_right_splitable ) {
        addPairHeap( heap, 
                     pair.get_left(),
                     pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left(),
                     pair.get_right()->get_right() );
        return;
    }
}


void   TreeDiamAlg2d::compute_by_heap( double  eps )
{
    heap_pairs_2d  heap;
    int  heap_limit, heap_delta;
    
    heap_limit = HEAP_LIMIT;
    heap_delta = HEAP_DELTA;

    FS2dTreeNode  * root = tree.getRoot();
    
    computeExtremePair( tree.getPoints(), root->getBBox().getLongestDim(),
                        pair_diam );
    
    FS2dPair  root_pair;
    
    root_pair.init( root, root );
    
    heap.push( root_pair );
    //queue_a.pushx( root_pair );
    
    int  count =0;
    while  ( heap.size() > 0 ) {
        //intf( "%d: curr_diam: %g\n", count++, pair_diam.distance );
        FS2dPair  pair = heap.top();
        if  ( ( count & 0x3ff ) == 0 ) {
            if  ( ((int)heap.size()) > heap_limit ) {
                threshold_brute *= 2;
                printf( "threshold_brute: %d\n", threshold_brute );
                heap_limit += heap_delta;
            } 
        }
        split_pair( pair, heap, eps );
        count++;
        
        heap.pop();
    }
}

void  computeExtremePairDir2d( const ArrPoint2dPtr  & arr,
                               const Point2d  & dir,
                               PointPair2d  & pp )
{
    real  min_dot, max_dot, prod;
    int  ind_min_dot, ind_max_dot;

    pp.init( *(arr[ 0 ]) );
    min_dot = max_dot = dir.dot_prod( *(arr[ 0 ]) );
    ind_min_dot = ind_max_dot = 0;

    for  ( int  ind = 1; ind < arr.size(); ind++ ) {
        prod = dir.dot_prod( *(arr[ ind ]) );
        if  ( prod < min_dot ) {
            min_dot = prod;
            ind_min_dot = ind;
        }
        if  ( prod > max_dot ) {
            max_dot = prod;
            ind_max_dot = ind;
        }
    }

    pp.init( *(arr[ ind_min_dot ]), *(arr[ ind_max_dot ]) );
}


//----------------------------------------------------------------------
// TreeDiamPAlg -
// fair split tree that uses pointers
//----------------------------------------------------------------------

class  heap_pairs_p : public std::priority_queue<FSPPair, 
                      std::vector<FSPPair>, 
                      lt_fsppair>
{
};
//   heap_pairs;
class  FSPPairArray : public  Array<FSPPair> {};

void  TreeDiamPAlg::compute_by_level( double  eps ) 
{
    FSPPairArray  queue_a, queue_b;
    FSPPairArray  * p_queue_in, * p_queue_out;
    int  heap_limit, heap_delta;
 
    //printf( "xa\n" );
    queue_a.init( 100, 43432 );
    queue_b.init( 100, 43431 );
    
    FSPTreeNode  * root = tree.getRoot();

    //printf( "xb\n" );
    computeExtremePair( tree.getPoints(), 
                        points_num,
                        root->getBBox().getLongestDim(),
                        pair_diam );

    FSPPair  root_pair;
    
    root_pair.init( root, root );
    
    //printf( "xc\n" );
    queue_a.pushx( root_pair );
    
    heap_limit = HEAP_LIMIT;
    heap_delta = HEAP_DELTA;

    p_queue_in = &queue_a;
    p_queue_out = &queue_b;
    for  ( int  ind = 0; ind < 1000; ind++ ) {
        //printf( "ind: %d\n", ind );
        if   ( p_queue_in->isEmpty() ) {
            //printf( "breaking!\n" );
            break;
        }
        p_queue_out->reset();
        //printf( "xd\n" );
        if  ( p_queue_in->size() > heap_limit ) {
                threshold_brute *= 4;
                printf( "threshold_brute: %d\n", threshold_brute );
                heap_limit += heap_delta;
            }

        split_level( *p_queue_in, *p_queue_out, eps );
        
        exchange( p_queue_in, p_queue_out );
    }
    
}


void  TreeDiamPAlg::compute_by_heap( double  eps ) 
{
    heap_pairs_p  heap;
    int  heap_limit, heap_delta;
    
    heap_limit = HEAP_LIMIT;
    heap_delta = HEAP_DELTA;
    
    FSPTreeNode  * root = tree.getRoot();
    
    computeExtremePair( tree.getPoints(), points_num,
                        root->getBBox().getLongestDim(),
                        pair_diam );
    //diam = root->getBBox().getLongestEdge();
    
    FSPPair  root_pair;
    
    root_pair.init( root, root );
    
    heap.push( root_pair );
    //queue_a.pushx( root_pair );
    
    int  count =0;
    while  ( heap.size() > 0 ) {
        //printf( "%d: curr_diam: %g\n", count++, pair_diam.distance );
        FSPPair  pair = heap.top();
        split_pair( pair, heap, eps );
        if  ( ( count & 0x3ff ) == 0 ) {
            if  ( ((int)heap.size()) > heap_limit ) {
                threshold_brute *= 2;
                printf( "threshold_brute: %d\n", threshold_brute );
                heap_limit += heap_delta;
            } 
        }
        count++;

        heap.pop();
    }
}


void  TreeDiamPAlg::addPair( FSPPairArray  & queue_out, 
                            FSPTreeNode  * left,
                            FSPTreeNode  * right )
{
    const Point3d  & p( **(left->ptr_pnt_left()) );
    const Point3d  & q( **(right->ptr_pnt_left()) );
    
    //printf( "old_diam; %g\n", diam );
    pair_diam.update_diam( p, q );

    //diam = max( diam, pnt_distance( p, q ) );
    //printf( "new_diam; %g\n", diam );
    
    FSPPair  pair;
    
    pair.init( left, right );
    if  ( pair.maxDiam() <= pair_diam.distance )
        return;
    queue_out.pushx( pair );
}


void  TreeDiamPAlg::split_level( FSPPairArray  & queue_in,
                                FSPPairArray  & queue_out, 
                                double  eps )
{
    bool  f_is_left_splitable, f_is_right_splitable;
    
    for  ( int  ind = 0; ind < queue_in.size(); ind++ ) {
        FSPPair  & in( queue_in[ ind ] );
        if  ( in.maxDiam() <= (1.0 + eps) * pair_diam.distance )
            continue;
        
        if  ( in.isBelowThreshold( threshold_brute ) ) {
            in.directDiameter( tree, pair_diam );
            continue;
        }
        
        //printf( "\n---curr_diam: %g\n", diam );
        //in.dump();
        //printf( "-------------\n\n" );
        // Oh well, we need to split, if possible...
        f_is_left_splitable = in.get_left()->isSplitable();
        f_is_right_splitable = in.get_right()->isSplitable();
        assert( f_is_left_splitable ||  f_is_right_splitable );
        if ( f_is_left_splitable )
            tree.split_node( in.get_left() );
        if ( f_is_right_splitable )
            tree.split_node( in.get_right() );
        
        if  ( f_is_left_splitable 
              &&  f_is_right_splitable ) {
            addPair( queue_out, 
                     in.get_left()->get_left(),
                     in.get_right()->get_left() );
            addPair( queue_out, 
                     in.get_left()->get_left(),
                     in.get_right()->get_right() );
            // to avoid exponential blowup
            if  ( in.get_left() != in.get_right() )
                addPair( queue_out, 
                         in.get_left()->get_right(),
                         in.get_right()->get_left() );
            addPair( queue_out, 
                     in.get_left()->get_right(),
                     in.get_right()->get_right() );
            continue;
        }
        if  ( f_is_left_splitable ) {
            addPair( queue_out, 
                     in.get_left()->get_left(),
                     in.get_right() );
            addPair( queue_out, 
                     in.get_left()->get_right(),
                     in.get_right() );
            continue;
        }
        if  ( f_is_right_splitable ) {
            addPair( queue_out, 
                     in.get_left(),
                     in.get_right()->get_left() );
            addPair( queue_out, 
                     in.get_left(),
                     in.get_right()->get_right() );
            continue;
        }
    }
}


void  TreeDiamPAlg::split_pair( FSPPair  & pair,
                               heap_pairs_p  & heap, double  eps )
{
    bool  f_is_left_splitable, f_is_right_splitable;
    
    if  ( pair.maxDiam() <= ((1.0 + eps) * pair_diam.distance ))
        return;
    
    if  ( pair.isBelowThreshold( threshold_brute ) ) {
        pair.directDiameter( tree, pair_diam );
        return;
    }
    
    f_is_left_splitable = pair.get_left()->isSplitable();
    f_is_right_splitable = pair.get_right()->isSplitable();
    assert( f_is_left_splitable  ||  f_is_right_splitable );
    if ( f_is_left_splitable )
        tree.split_node( pair.get_left() );
    if ( f_is_right_splitable )
        tree.split_node( pair.get_right() );
    
    if  ( f_is_left_splitable 
          &&  f_is_right_splitable ) {
        addPairHeap( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left()->get_left(),
                     pair.get_right()->get_right() );
        // to avoid exponential blowup
        if  ( pair.get_left() != pair.get_right() )
            addPairHeap( heap, 
                         pair.get_left()->get_right(),
                         pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right()->get_right() );
        return;
    }
    if  ( f_is_left_splitable ) {
        addPairHeap( heap,
                     pair.get_left()->get_left(),
                     pair.get_right() );
        addPairHeap( heap, 
                     pair.get_left()->get_right(),
                     pair.get_right() );
        return;
    }
    if  ( f_is_right_splitable ) {
        addPairHeap( heap, 
                     pair.get_left(),
                     pair.get_right()->get_left() );
        addPairHeap( heap, 
                     pair.get_left(),
                     pair.get_right()->get_right() );
        return;
    }
}

/*
real   brute_diameter( const ArrPoint3d  & arr,
                       PointPair  & diam )
{
    diam.init( arr[ 0 ] );

    for  ( int  ind = 0; ind < ( arr.size() - 1 ); ind++ )
        for  ( int  jnd = ind + 1; jnd < arr.size(); jnd++ ) 
            diam.update_diam( arr[ ind ], arr[ jnd ] );

    return  diam.distance;
}*/

void  TreeDiamPAlg::addPairHeap( heap_pairs_p  & heap,
                             FSPTreeNode  * left,
                             FSPTreeNode  * right )
{
    const Point3d  & p( **(left->ptr_pnt_left()) );
    const Point3d  & q( **(right->ptr_pnt_left()) );

    pair_diam.update_diam( p, q );
    //printf( "old_diam; %g\n", diam );
    //diam = max( diam, pnt_distance( p, q ) );
    //printf( "new_diam; %g\n", diam );
    
    FSPPair  pair;
    
    pair.init( left, right );
    if  ( pair.maxDiam() <= pair_diam.distance )
        return;
    heap.push( pair );
}

/* TreeDiam.C - End of File ------------------------------------------*/
