/*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
 * algorithm.C -
 *     Algorithms for computing the diameter
 * 
 * Copyright 2000 Sariel Har-Peled (ssaarriieell@cs.uiuc.edu)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * Code is based on the paper:
 *   A Practical Approach for Computing the Diameter of a Point-Set.
 *   Sariel Har-Peled (http://www.uiuc.edu/~sariel)
\*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*/

#include  <iostream>
#include  <assert.h>
#include  <stdlib.h>
#include  <stdio.h>
#include  <time.h>
#include  <math.h>

#include  <queue>

#include  <qapplication.h>
#include  <qgl.h>

//#include  "dmalloc.h"

//#include  "scape/vector3d.h"
//#include  "line.h"
#include  "globjwin.h"

#include  "_generic.h"
#include  "sarray.h"
#include  "real.h"
#include  "point.h"
#include  "bbox.h"
#include  "ftree.h"
#include  "ftree2d.h"
#include  "ftreeptr.h"
#include  "TreeDiam.h"
#include  "proj_points.h"
#include  "misc.h"
#include  "obb.h"

#include  "algorithm.h"
#include  "gdiam.h"


/*--- Constants ---*/

//------------------------------------------------------------------
// DiameterAlgorithm - Generic class for algorithms computing diam
//------------------------------------------------------------------
void   DiameterAlgorithm::dump()
{
    printf( "%-25s  %10.3f   %10.3g  eps: %10.4g\n",                   
            getID(),
            time,
            diam.distance,
            eps );
}


void   DiameterAlgorithm::dump_to_file( FILE  * fl )
{
    fprintf( fl, "%10.2f\n",                   
             time );
}


DiameterAlgorithm  * BruteDiameter::CreateAndCompute( ArrPoint3d  & arr )
{
    BruteDiameter  * pDiam = new BruteDiameter;

    pDiam->eps = 0;
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


char   * BruteDiameter::getID()
{
    static char  * name = "Brute force diameter";

    return  name;
}

void   BruteDiameter::compute( const ArrPoint3d  & arr )
{
    Timer  tm;

    diam.init( arr[ 0 ] );

    tm.start();

    brute_diameter( arr, diam );

    tm.end();
    time = tm.seconds();
}

//------------------------------------------------------------------
// BruteDiameterLocal
//    Compute diameter in a brute force manner - using cache locality
//    to improve speed.
//------------------------------------------------------------------

char   * BruteDiameterLocal::getID()
{
    static char  * name = "Brute force diam local";

    return  name;
}


DiameterAlgorithm  * BruteDiameterLocal::CreateAndCompute( 
                                                 ArrPoint3d  & arr )
{
    BruteDiameterLocal  * pDiam = new BruteDiameterLocal;

    pDiam->eps = 0;
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}

    
void   BruteDiameterLocal::compute( const ArrPoint3d  & arr )
{
    Timer  tm;

    diam.init( arr[ 0 ] );
    tm.start();

    brute_diameter_local( arr, diam );

    tm.end();
    time = tm.seconds();    
}


//------------------------------------------------------------------
// QuadDiameterByLevel
//------------------------------------------------------------------
char   * QuadDiameterByLevel::getID()
{
    static char  * name = "FairSplit using levels";

    return  name;
}


void   QuadDiameterByLevel::compute( const ArrPoint3d  & arr )
{
    Timer  tm;

    tm.start();

    p_alg = new TreeDiamAlg( arr );
    p_alg->compute_by_level( eps );
    tm.end();

    diam = p_alg->getDiameter();

    time = tm.seconds();   
    p_alg->term();
    delete  p_alg;
    p_alg = NULL;
}


DiameterAlgorithm  * QuadDiameterByLevel::CreateAndCompute( 
                         ArrPoint3d  & arr,
                         double  _eps )
{
    QuadDiameterByLevel   * pDiam = new QuadDiameterByLevel;

    pDiam->eps = _eps;

    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}

//------------------------------------------------------------------
// QuadDiameterPtrByLevel
//------------------------------------------------------------------
char   * QuadDiameterPtrByLevel::getID()
{
    static char  * name = "FairSplitPtr using levels";

    return  name;
}


void   QuadDiameterPtrByLevel::compute( const ArrPoint3d  & arr )
{
    Timer  tm;

    tm.start();

    //printf( "a\n" );
    p_alg = new TreeDiamPAlg( arr );
    //printf( "b\n" );-
    p_alg->compute_by_level( eps );
    //printf( "c\n" );
    tm.end();

    diam = p_alg->getDiameter();

    time = tm.seconds();    
    p_alg->term();
    delete  p_alg;
    p_alg = NULL;
}

DiameterAlgorithm  * QuadDiameterPtrByLevel::CreateAndCompute( 
                         ArrPoint3d  & arr,
                         double  _eps )
{
    QuadDiameterPtrByLevel   * pDiam = new QuadDiameterPtrByLevel;

    pDiam->eps = _eps;

    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


//------------------------------------------------------------------
// QuadDiameterByDir
//------------------------------------------------------------------
char   * QuadDiameterByDir::getID()
{
    static char  * name = "FairSplit by direction";

    return  name;
}


void   QuadDiameterByDir::compute( const ArrPoint3d  & arr )
{
    Timer  tm;

    tm.start();

    p_alg = new TreeDiamAlg( arr );
    p_alg->compute_by_dir( eps );
    tm.end();

    diam = p_alg->getDiameter();

    time = tm.seconds();    
    p_alg->term();
    delete  p_alg;
    p_alg = NULL;
}

DiameterAlgorithm  * QuadDiameterByDir::CreateAndCompute( 
                         ArrPoint3d  & arr,
                         double  _eps )
{
    QuadDiameterByDir   * pDiam = new QuadDiameterByDir;

    pDiam->eps = _eps;

    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}



//------------------------------------------------------------------
// QuadDiameterByHeap
//------------------------------------------------------------------
char   * QuadDiameterByGPtrHeap::getID()
{
    static char  * name = "FS gdiam ptrheap";

    return  name;
}


void   QuadDiameterByGPtrHeap::compute( const ArrPoint3d  & arr )
{
    Timer  tm;
    gdiam_real  * p_arr, * p_start;
    //double  num;

    tm.start();

    p_arr = (gdiam_real *)malloc( sizeof( gdiam_real ) * arr.size() * 3 );
    p_start = p_arr;
    for  ( int  ind = 0; ind < arr.size(); ind++ ) {
        p_arr[ 0 ] = arr[ ind ][ 0 ];
        p_arr[ 1 ] = arr[ ind ][ 1 ];
        p_arr[ 2 ] = arr[ ind ][ 2 ];
        p_arr += 3;
    }

    GPointPair  pair;

    pair = gdiam_approx_diam_pair( p_start, arr.size(), eps );
    //gdiam_approx_const_mvbb( p_start, arr.size(), eps );
    
    diam.init( arr[ (pair.p - p_start) / 3 ],
               arr[ (pair.q - p_start) / 3 ] );

    free( p_start );
    
    tm.end();

    time = tm.seconds();    
}


DiameterAlgorithm  * QuadDiameterByGPtrHeap::CreateAndCompute( 
                             ArrPoint3d  & arr, double  _eps )
{
    QuadDiameterByGPtrHeap  * pDiam;

    pDiam = new QuadDiameterByGPtrHeap;

    pDiam->eps = _eps;

    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


//------------------------------------------------------------------
// QuadDiameterByHeapUDMSample
//------------------------------------------------------------------
char   * QuadDiameterByGPtrHeapUDMSample::getID()
{
    static char  * name = "FS(UDM) gdiam ptrheap";

    return  name;
}



void   QuadDiameterByGPtrHeapUDMSample::compute( const ArrPoint3d  & arr )
{
    Timer  tm;
    gdiam_real  * p_arr, * p_start;
    //double  num;

    tm.start();

    p_arr = (gdiam_real *)malloc( sizeof( gdiam_real ) * arr.size() * 3 );
    p_start = p_arr;
    for  ( int  ind = 0; ind < arr.size(); ind++ ) {
        p_arr[ 0 ] = arr[ ind ][ 0 ];
        p_arr[ 1 ] = arr[ ind ][ 1 ];
        p_arr[ 2 ] = arr[ ind ][ 2 ];
        p_arr += 3;
    }

    GPointPair  pair;

    pair = gdiam_approx_diam_pair_UDM( p_start, arr.size(), eps );
    //gdiam_approx_const_mvbb( p_start, arr.size(), eps );
    
    diam.init( arr[ (pair.p - p_start) / 3 ],
               arr[ (pair.q - p_start) / 3 ] );

    free( p_start );
    
    tm.end();

    time = tm.seconds();    
}

DiameterAlgorithm  * QuadDiameterByGPtrHeapUDMSample::CreateAndCompute( 
                             ArrPoint3d  & arr, double  _eps )
{
    QuadDiameterByGPtrHeapUDMSample  * pDiam;

    pDiam = new QuadDiameterByGPtrHeapUDMSample;

    pDiam->eps = _eps;

    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


//------------------------------------------------------------------
// QuadDiameterByHeap
//------------------------------------------------------------------
char   * QuadDiameterByHeap::getID()
{
    static char  * name = "FairSplit using heap";

    return  name;
}


void   QuadDiameterByHeap::compute( const ArrPoint3d  & arr )
{
    Timer  tm;

    tm.start();

    p_alg = new TreeDiamAlg( arr );
    p_alg->compute_by_heap( eps );
    tm.end();

    diam = p_alg->getDiameter();

    time = tm.seconds();    
    p_alg->term();
    delete  p_alg;
    p_alg = NULL;
}

DiameterAlgorithm  * QuadDiameterByHeap::CreateAndCompute( 
                             ArrPoint3d  & arr, double  _eps )
{
    QuadDiameterByHeap  * pDiam = new QuadDiameterByHeap;

    pDiam->eps = _eps;

    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}

//------------------------------------------------------------------
// QuadDiameterByHeapWSPD
//------------------------------------------------------------------
char   * QuadDiameterByHeapWSPD::getID()
{
    static char  * name = "FairSplit using heap WSPD";

    return  name;
}


void   QuadDiameterByHeapWSPD::compute( const ArrPoint3d  & arr )
{
    Timer  tm;

    tm.start();


    p_alg = new TreeDiamAlg( arr );
    p_alg->compute_by_heap_wspd( eps );
    tm.end();

    diam = p_alg->getDiameter();

    time = tm.seconds();    
    p_alg->term();
    delete  p_alg;
    p_alg = NULL;
}

DiameterAlgorithm  * QuadDiameterByHeapWSPD::CreateAndCompute( 
                             ArrPoint3d  & arr, double  _eps )
{
    QuadDiameterByHeapWSPD  * pDiam = new QuadDiameterByHeapWSPD;

    pDiam->eps = _eps;

    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


//------------------------------------------------------------------
// DiameterAPBBox -
//    Compute the largest pair from the bounding box...
//------------------------------------------------------------------
DiameterAlgorithm  * DiameterAPBBox::CreateAndCompute( ArrPoint3d  & arr )
{
    DiameterAPBBox  * pDiam = new DiameterAPBBox;

    pDiam->eps = sqrt(3);
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}

char   * DiameterAPBBox::getID()
{
    static char  * name = "Diameter from BBox";

    return  name;
}


void   DiameterAPBBox::compute( const ArrPoint3d  & arr )
{
    Timer  tm;
    PointPair  pp0, pp1, pp2;
    
    diam.init( arr[ 0 ] );
    tm.start();
    
    computeExtremePairAllDim( arr, pp0, pp1, pp2 );

    diam.update_diam( pp0 );
    diam.update_diam( pp1 );
    diam.update_diam( pp2 );

    tm.end();
 
    time = tm.seconds();    
}


//------------------------------------------------------------------
// DiameterDirections -
//     Approx diameter by projecting the points into alot of
//   directions
//------------------------------------------------------------------
DiameterAlgorithm  * DiameterDirections::CreateAndCompute( ArrPoint3d  & arr,
                         double  _eps )
{
    DiameterDirections   * pDiam = new DiameterDirections;

    pDiam->eps = _eps;
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}

char   * DiameterDirections::getID()
{
    static char  * name = "Diametr from directions";

    return  name;
}


void   DiameterDirections::compute( const ArrPoint3d  & arr )
{
    Timer  tm;
    int  alpha_limit, beta_limit;
    real  delta, alpha, beta;

    diam.init( arr[ 0 ] );
    tm.start();
    
    delta = sqrt( eps );
    alpha_limit = (int)(2.0 * 3.1415926535 / delta);
    beta_limit = (int)(3.1415926535 / delta);
    
    alpha = 0.0;
    for  ( int  ind_alpha = 0; ind_alpha < alpha_limit; 
           ind_alpha++, alpha += delta ) {
        beta = -3.1415926535 / 2.0;
        for  ( int  ind_beta = 0; ind_beta < beta_limit; 
               ind_beta++, beta += delta ) {
            Point3d  pnt;

            pnt.setCoord( 0, cos( beta ) * sin( alpha ) );
            pnt.setCoord( 1, cos( beta ) * cos( alpha ) );
            pnt.setCoord( 2, sin( beta ) );

            PointPair  pp;
            computeExtremePairDir( arr, pnt, pp );
            diam.update_diam( pp );
        }
    }
    
    tm.end();
    time = tm.seconds();    
}


//------------------------------------------------------------------
// DiameterGrid -
//     We filter out points by using a grid,
//------------------------------------------------------------------

class  GridCleaner
{
public:
    ArrPoint3d   arr, out_arr;
    Point3d  left_bottom;
    int  side_size, overall_size;
    real  x_delta, y_delta, z_delta;
    Point3d  ** shafts_bottom, ** shafts_top;

    ~GridCleaner() {
        arr.term();
        out_arr.term();
        free( shafts_bottom );
        free( shafts_top );
        shafts_bottom = shafts_top = NULL;
    }

    void  init( const ArrPoint3d  & _arr, 
                double   eps );
    void  map_xy( Point3d  pnt, int  & x, int  & y ) {
        x = (int)( ( pnt[ 0 ] - left_bottom[ 0 ] ) / x_delta );
        y = (int)( ( pnt[ 1 ] - left_bottom[ 1 ] ) / y_delta );
    }
    int  cell( Point3d  pnt ) {
        int  x, y;

        map_xy( pnt, x, y );
        //printf( "x: %d, y: %d  size_side: %d\n", x, y,
        //        side_size );
        if  ( y == side_size )
            y--;
        return  y * side_size + x;
    }
    void   register_point( Point3d  & pnt ) {
        int  ind = cell( pnt );

        if  ( ( ind < 0 )  ||  ( ind >= overall_size ) ) {
            printf( "ind: %d, overall size: %d\n",
                    ind, overall_size );
            assert( (0 <= ind )  &&  ( ind < overall_size ) );
        }

        if   ( ( shafts_bottom[ ind ] == NULL )
               ||  ( (*(shafts_bottom[ ind ]))[ 2 ] < pnt[ 2 ] ) )
            shafts_bottom[ ind ] = &pnt;

        if   ( ( shafts_top[ ind ] == NULL )
               ||  ( (*(shafts_top[ ind ]))[ 2 ] > pnt[ 2 ] ) )
            shafts_top[ ind ] = &pnt;
    }
};


void  GridCleaner::init( const ArrPoint3d  & _arr, 
                         double   eps )
{ 
    arr.init( _arr.size(), 43543 );

    for  ( int  ind = 0; ind <  _arr.size(); ind++ ) 
        arr.pushx( _arr[ ind ] );

    BBox3d  bb;

    pnt_array_bound( arr, bb );
    
    left_bottom.setCoord( 0, bb.min_coord( 0 ) );
    left_bottom.setCoord( 1, bb.min_coord( 1 ) );
    left_bottom.setCoord( 2, bb.min_coord( 2 ) );

    side_size = (int)( 1.0 + sqrt( 6.0 ) / eps );
    x_delta = (1.00000001 * bb.length_dim( 0 ) ) / (double)side_size;
    y_delta = (1.00000001 * bb.length_dim( 1 ) ) / (double)side_size;
    z_delta = (1.00000001 * bb.length_dim( 2 ) ) / (double)side_size;
 
    overall_size = side_size * side_size;

    shafts_bottom = (Point3d **)malloc( sizeof( Point3d * ) * overall_size );
    assert( shafts_bottom != NULL );
    memset( shafts_bottom, 0, sizeof( Point3d * ) * overall_size  );

    shafts_top = (Point3d **)malloc( sizeof( Point3d * ) * overall_size );
    assert( shafts_top != NULL );
    memset( shafts_top, 0, sizeof( Point3d * ) * overall_size  );
    
    for  ( int  ind = 0; ind <  arr.size(); ind++ ) 
        register_point( arr[ ind ] );

    out_arr.init( 100, 45344354 );
    for  ( int  ind = 0; ind < overall_size; ind++ ) { 
        if  ( ( shafts_bottom[ ind ] != NULL )
              &&  ( shafts_top[ ind ] != NULL ) ) {
            out_arr.pushx( *( shafts_bottom[ ind ] ) );
            if  ( shafts_bottom[ ind ] != shafts_top[ ind ] )
                out_arr.pushx( *( shafts_top[ ind ] ) );
        } else {
            if  ( shafts_bottom[ ind ] != NULL )
                out_arr.pushx( *( shafts_bottom[ ind ] ) );
            if  ( shafts_top[ ind ] != NULL )
                out_arr.pushx( *( shafts_top[ ind ] ) );
            
        }
    }
}


DiameterAlgorithm  * DiameterGrid::CreateAndCompute( ArrPoint3d  & arr,
                                                     double  _eps )
{
    DiameterGrid  * pDiam = new DiameterGrid;

    pDiam->eps = _eps;
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


char   * DiameterGrid::getID() 
{
    static char  * name = "Diameter grid";

    return  name;
}


void   DiameterGrid::compute( const ArrPoint3d  & arr )
{
    Timer  tm;
    GridCleaner  grid;
    
    diam.init( arr[ 0 ] );

    tm.start();

    grid.init( arr, eps / 2.0 );
    DiameterAlgorithm  * pDiamHeap;

    printf( "grid cleaning\n"
            "\tin_size : %d\n"
            "\tout_size: %d\n", 
            arr.size(), grid.out_arr.size() );
    pDiamHeap = QuadDiameterByLevel::CreateAndCompute( grid.out_arr, 
                              
                         eps / 2.0 );
    
    diam = pDiamHeap->diam;

    tm.end();
    time = tm.seconds();    
}


//------------------------------------------------------------------
// DiameterGridQuadDir -
//     We filter out points by using a grid, and then use fiar 
// split with direcitons
//------------------------------------------------------------------


DiameterAlgorithm  * DiameterGridQuadDir::CreateAndCompute( ArrPoint3d  & arr,
                                                            double  _eps )
{
    DiameterGridQuadDir  * pDiam = new DiameterGridQuadDir;

    pDiam->eps = _eps;
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


char   * DiameterGridQuadDir::getID() 
{
    static char  * name = "Diameter grid->FSTree dir";

    return  name;
}


void   DiameterGridQuadDir::compute( const ArrPoint3d  & arr )
{
    Timer  tm;
    GridCleaner  grid;
    
    diam.init( arr[ 0 ] );

    tm.start();

    grid.init( arr, eps / 2.0 );
    DiameterAlgorithm  * pDiamHeap;

    //printf( "\tgrid cleanring in_size: %d\nout_size: %d\n", 
    //        arr.size(), grid.out_arr.size() );
    pDiamHeap = QuadDiameterByDir::CreateAndCompute( grid.out_arr, eps/2 );
    
    diam = pDiamHeap->diam;

    tm.end();
    time = tm.seconds();    
}


//------------------------------------------------------------------
// ModifiedDiameterChan - 
//     Describes the algorithm due to Timothy Chan Socg2000
//     On the 2d level we use cleaup & fiar split tree to get
//   better resuls.
//------------------------------------------------------------------

DiameterAlgorithm  * DiameterModChan::CreateAndCompute(
                                              ArrPoint3d  & arr,
                                              double  _eps )
{
    DiameterModChan  * pDiam = new DiameterModChan;

    pDiam->eps = _eps;
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


char   * DiameterModChan::getID()
{
    static char  * name = "Chan's mod algorithm";

    return  name;
}


void   DiameterModChan::compute( const ArrPoint3d  & arr )
{
    Timer  tm;
    GridCleaner  grid;
    double  delta, alpha, alpha_limit;

    diam.init( arr[ 0 ] );

    tm.start();

    //for  ( int  ind = 0; ind < arr.size(); ind++ ) 
    //    printf( "___id: %d\n", arr[ ind ].getID() );
        
    grid.init( arr, eps / 3.0 );


    //for  ( int  ind = 0; ind < grid.out_arr.size(); ind++ ) 
    //    printf( "x___id: %d\n", grid.out_arr[ ind ].getID() );

    //printf( "in_size: %d\nout_size: %d\n", 
    //        arr.size(), grid.out_arr.size() );
    
    delta = sqrt( eps / 3.0 );
    alpha_limit = (int)(3.1415926535 / delta);

    alpha = 0.0;
    for  ( int  ind_alpha = 0; ind_alpha < alpha_limit; 
           ind_alpha++, alpha += delta ) {
        ProjPoints  prj;
        Point3d  dir;
        ArrPoint2dPtr  out_arr;

        dir[ 0 ] = sin( alpha );
        dir[ 1 ] = cos( alpha );
        dir[ 3 ] = 0;
        
        prj.init( grid.out_arr, dir );

        // Figure out the exact constants...
        prj.grid_clean( 1 + (int)(3.0 * sqrt(2) / eps),
                        out_arr );
        
        TreeDiamAlg2d  * p_alg;

        p_alg = new TreeDiamAlg2d( out_arr );
        p_alg->compute_by_heap( eps / 2.0 );
        
        PointPair2d  diam2d = p_alg->getDiameter();

        PointPair   new_diam;
        new_diam.init( arr[ diam2d.p.getID() ],
                       arr[ diam2d.q.getID() ] );
        //printf( "id: %d, %d\n", diam2d.p.getID(), diam2d.q.getID() );

        diam.update_diam( new_diam );
        p_alg->term();
        delete  p_alg;
    }

    tm.end();
    time = tm.seconds();    
}


//------------------------------------------------------------------
// DiameterChan - 
//     Describes the algorithm due to Timothy Chan Socg2000
//------------------------------------------------------------------

DiameterAlgorithm  * DiameterChan::CreateAndCompute(
                                              ArrPoint3d  & arr,
                                              double  _eps )
{
    DiameterChan  * pDiam = new DiameterChan;

    pDiam->eps = _eps;
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


char   * DiameterChan::getID()
{
    static char  * name = "Chan's algorithm";

    return  name;
}


void   DiameterChan::compute( const ArrPoint3d  & arr )
{
    Timer   tm;
    GridCleaner   grid;
    double   delta, alpha, beta;
    int   alpha_limit, beta_limit;

    diam.init( arr[ 0 ] );

    tm.start();

    //for  ( int  ind = 0; ind < arr.size(); ind++ ) 
    //    printf( "___id: %d\n", arr[ ind ].getID() );
        
    grid.init( arr, eps / 3.0 );


    //for  ( int  ind = 0; ind < grid.out_arr.size(); ind++ ) 
    //    printf( "x___id: %d\n", grid.out_arr[ ind ].getID() );

    //printf( "in_size: %d\nout_size: %d\n", 
    //        arr.size(), grid.out_arr.size() );
    
    delta = sqrt( eps / 3.0 );
    alpha_limit = (int)(3.1415926535 / delta);
    beta_limit = (int)(3.1415926535 / delta);

    alpha = 0.0;
    for  ( int  ind_alpha = 0; ind_alpha < alpha_limit; 
           ind_alpha++, alpha += delta ) {
        ProjPoints  prj;
        Point3d  dir;
        ArrPoint2dPtr  out_arr;

        dir[ 0 ] = sin( alpha );
        dir[ 1 ] = cos( alpha );
        dir[ 3 ] = 0;
        
        prj.init( grid.out_arr, dir );

        // Figure out the exact constants...
        prj.grid_clean( 1 + (int)(3.0 * sqrt(2) / eps),
                        out_arr );        
        
        beta = 0;
        for  ( int  ind_beta = 0; ind_beta < beta_limit; 
               ind_beta++, beta += delta ) {
            Point2d  pnt2d;

            pnt2d[ 0 ] = sin( beta );
            pnt2d[ 1 ] = cos( beta );
            
            PointPair2d  diam_tmp;
            //find_extreme_diameter_in_dir( out_arr, pnt2d, diam_tmp );
            computeExtremePairDir2d( out_arr, pnt2d, diam_tmp );

            PointPair   new_diam;
            new_diam.init( arr[ diam_tmp.p.getID() ],
                           arr[ diam_tmp.q.getID() ] );
            diam.update_diam( new_diam );
        }
    }

    tm.end();
    time = tm.seconds();    
}


//------------------------------------------------------------------
// DiameterPCA - 
//       Diameter using Principal component analysis.
//------------------------------------------------------------------


DiameterAlgorithm  * DiameterPCA::CreateAndCompute( ArrPoint3d  & arr )
{
    DiameterPCA  * pDiam = new DiameterPCA;

    pDiam->eps = sqrt(3);
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


char   * DiameterPCA::getID()
{
    static char  * name = "PCA";

    return  name;
}


void   DiameterPCA::compute( const ArrPoint3d  & arr )
{
    Timer   tm;
    Point3d  base_pnt, side_0, side_1, side_2;

    diam.init( arr[ 0 ] );

    tm.start();
    
    MakeCovarOBB( arr, base_pnt, side_0, side_1, side_2 );
    
    PointPair  pp_0, pp_1, pp_2;
    
    diam.init( arr[ 0 ] );
    tm.start();

    computeExtremePairDirs( arr, 
                            side_0, side_1, side_2, 
                            pp_0, pp_1, pp_2 );
    diam.update_diam( pp_0 );
    diam.update_diam( pp_1 );
    diam.update_diam( pp_2 );
        /*
    computeExtremePairDir( arr, side_0, pp );
    diam.update_diam( pp );

    computeExtremePairDir( arr, side_1, pp );
    diam.update_diam( pp );

    computeExtremePairDir( arr, side_2, pp );
    diam.update_diam( pp );
        */

    tm.end();
    //printf( "PCA: \n" );
    //side_0.print();
    //printf( "\n" );
    //side_1.print();
    ////printf( "\n" );
    //side_2.print();
    //printf( "\n" );
 
    time = tm.seconds();    
}


//------------------------------------------------------------------
// DiameterMassPCA - 
//       Diameter using Principal component analysis.
//------------------------------------------------------------------


DiameterAlgorithm  * DiameterMassPCA::CreateAndCompute( ArrPoint3d  & arr )
{
    DiameterMassPCA  * pDiam = new DiameterMassPCA;

    pDiam->eps = sqrt(3);
    pDiam->compute( arr );

    return  pDiam;  //(DiameterAlgorithm *)
}


char   * DiameterMassPCA::getID()
{
    static char  * name = "Mass PCA";

    return  name;
}


void   DiameterMassPCA::compute( const ArrPoint3d  & arr )
{
    Timer   tm;
    //Point3d  base_pnt, side_0, side_1, side_2;
    FILE  * fl;
    Point3d  base_pnt, side_0, side_1, side_2;
    int  args;
    double  x, y, z;
    double  mCoVars[3][3];

    printf( "void   DiameterMassPCA::compute( const ArrPoint3d  & arr )\n" );
    tm.start();

    fl = fopen( "points_out.txt", "wt" );
    assert( fl != NULL );

    fprintf( fl, "3\n%d", arr.size() );
    for  ( int  ind = 0; ind < arr.size(); ind++ ) {
        fprintf( fl, "\n%.20f %.20f %.20f ",
                 arr[ ind ][ 0 ],
                 arr[ ind ][ 1 ],
                 arr[ ind ][ 2 ] );
    }
    fprintf( fl, "\n" );
    fclose( fl );
    system( "./qhuller points_out.txt" );

    // Read the Mass PCA results
    fl = fopen( "out2.txt", "rt" );
    assert( fl != NULL );


    args = fscanf( fl, "%lg %lg %lg\n", &x, &y, &z );
    assert( args == 3 );

    //Point3d  mass_center( x, y, z );

    args = fscanf( fl, "%lg %lg %lg\n", &x, &y, &z );
    assert( args == 3 );
                         
    mCoVars[ 0 ][ 0 ] = x;
    mCoVars[ 0 ][ 1 ] = y;
    mCoVars[ 0 ][ 2 ] = z;

    args = fscanf( fl, "%lg %lg %lg\n", &x, &y, &z );
    assert( args == 3 );
    mCoVars[ 1 ][ 0 ] = x;
    mCoVars[ 1 ][ 1 ] = y;
    mCoVars[ 1 ][ 2 ] = z;

    args = fscanf( fl, "%lg %lg %lg\n", &x, &y, &z );
    assert( args == 3 );
    mCoVars[ 2 ][ 0 ] = x;
    mCoVars[ 2 ][ 1 ] = y;
    mCoVars[ 2 ][ 2 ] = z;

    fclose( fl );

    double  max_entry;

    max_entry = 0;
    for  ( int  ind = 0; ind < 3; ind++ ) 
        for  ( int  jnd = 0; jnd < 3; jnd++ )
            max_entry = max( max_entry, fabs( mCoVars[ ind ][ jnd ] ) );

    if  ( max_entry != 0.0 ) {
        for  ( int  ind = 0; ind < 3; ind++ ) 
            for  ( int  jnd = 0; jnd < 3; jnd++ ) {
                mCoVars[ ind ][ jnd ] = mCoVars[ ind ][ jnd ] / max_entry;
                if  ( fabs( mCoVars[ ind ][ jnd ] ) < 1e-8 ) 
                    mCoVars[ ind ][ jnd ] = 0;
            }
    }


    //printf( "\nCenter of mass: " );
    //mass_center.print();
    //printf( "\n" );

    /*
    for  ( int  ind = 0; ind < 3; ind++ ) {
        printf( "--- [ " );
        for  ( int  jnd = 0; jnd < 3; jnd++ )
            printf( "%g ", mCoVars[ ind ][ jnd ] );
        printf( "]\n" );
    }
    */

    Point3d  vec1, vec2, vec3;
    EigenSymetric3vec( mCoVars, vec1, vec2, vec3 );

    vec1.normalize();
    vec2.normalize();
    vec3.normalize();
    //extractOrthonormalBase( vec1, vec2, vec3 );

    /*
    printf( "\n\n----------------- directions \n" );
    vec1.print();
    printf( "\n" );
    vec2.print();
    printf( "\n" );
    vec3.print();
    printf( "\n" );
    */
    computeParallelogram( arr, vec1, vec2, vec3,
                          base_pnt, side_0, side_1, side_2 );

    diam.init( arr[ 0 ] );
    /*
    printf( "sides...\n" );
    side_0.print();
    printf( "\n" );
    side_1.print();
    printf( "\n" );
    side_2.print();
    printf( "\n" );
    printf( "\n" );
    */
    PointPair  pp;

    computeExtremePairDir( arr, side_0, pp );
    diam.update_diam( pp );

    computeExtremePairDir( arr, side_1, pp );
    diam.update_diam( pp );

    computeExtremePairDir( arr, side_2, pp );
    diam.update_diam( pp );

    tm.end();

    time = tm.seconds();    
}

#define  LINE_SIZE  1024

static void   get_line( FILE  * fl, char  * line )
{
    char  * out;

    out = fgets( line, 1024, fl );
    assert( out != NULL );
}

static  bool  prefix_eq( char  * line, char  * prefix )
{
    while  ( *prefix ) {
        if  ( *line != *prefix )
            return  false;
        line++;
        prefix++;
    }
    return  true;
}


static  bool  suffix_eq( const char  * line, const char  * suffix )
{
    int  len_line, len_suffix;

    len_line = strlen( line );
    len_suffix = strlen( suffix );
    if  ( len_suffix == 0 )
        return  true;
    if  ( len_suffix > len_line )
        return  false;

    return ( strcmp( line + len_line - len_suffix, suffix ) == 0 );
}


void  read_poly_file( ArrPoint3d  & arr,
                      const char  * file_name )
{
    FILE  * fl;
    char  line[ LINE_SIZE ];
    int  vertices, args;

    fl = fopen( file_name, "rt" );
    if  ( fl == NULL ) {
        fprintf( stderr, "Unable to open file: [%s]\n",
                 file_name );
        exit( -1 );
    }

    while  ( ( ! feof( fl ) )  && 
             ( ! prefix_eq( line, "element vertex" ) ) ) 
        get_line( fl, line );
    if  ( feof( fl ) ) {
        fprintf( stderr, "Vertex num info not found!\n" );
        exit( -1 );
    }
    args = sscanf( line, "element vertex %d", &vertices );
    if  ( args < 1 ) {
        fprintf( stderr, "Vertex num info non decipherable!\n" );
        exit( -1 );
    }
    assert( vertices > 0 );
    
    // skip till end of header
    while  ( ( ! feof( fl ) )  && 
             ( ! prefix_eq( line, "end_header" ) ) ) 
        get_line( fl, line );
    if  ( feof( fl ) ) {
        fprintf( stderr, "header termination not found!\n" );
        exit( -1 );
    }

    arr.init( vertices + 10, 3454 );

    // read the vertices
    for  ( int  ind = 0; ind < vertices; ind++ ) {
        double  x, y, z;

        get_line( fl, line );
        if  ( feof( fl ) ) {
            fprintf( stderr, "unexpected end of file!\n" );
            exit( -1 );
        }
        args = sscanf( line, "%lg %lg %lg", &x, &y, &z );
        Point3d  pnt;
        pnt.setCoord( 0, x  );
        pnt.setCoord( 1, y );
        pnt.setCoord( 2, z );

        arr.pushx( pnt );
    }
}

PointPair  * tester_inner( ArrPoint3d   * p_arr,
                           const char  * description )
{
    DiameterAlgorithm  * alg;

    if  ( p_arr == NULL )
        return  NULL;

    TesterDiamAlg  tester;

    printf( "\n\n" );
    printf( "Performing extended tests. Thist might take a while...\n" );
    printf( "Points: %d\n", p_arr->size() );
    tester.init();
    tester.points_num = p_arr->size();

    //-compute_mvbb( *p_arr );

    //printf( "XXXXQuad by heap gdiam ptr ...\n" );
    //tester.push( alg = QuadDiameterByGPtrHeap::CreateAndCompute( 
    //                                               *p_arr, 0.0 ) );
    //-----------------------------------------------------------------
    // 0.0
    //-----------------------------------------------------------------
    printf( "Quad by heap...\n" );
    tester.push( alg = QuadDiameterByHeap::CreateAndCompute( *p_arr, 0.0 ) );

    printf( "Quad by heap gdiam ptr ...\n" );
    tester.push( alg = QuadDiameterByGPtrHeap::CreateAndCompute( 
                                                   *p_arr, 0.0 ) );
    //printf( "Quad by heap gdiam ptr UDM ...\n" );
    //tester.push( alg = QuadDiameterByGPtrHeapUDMSample::CreateAndCompute( 
    //                                               *p_arr, 0.0 ) );

    printf( "Quad by heap WSPD...\n" );
    tester.push( alg = QuadDiameterByHeapWSPD::CreateAndCompute( *p_arr,
                                                                 0.0 ) );
    printf( "Quad by level_ptr 0.0...\n" );
    tester.push( QuadDiameterPtrByLevel::CreateAndCompute( *p_arr, 0.0 ) );

    printf( "Quad by level...\n" );
    tester.push( QuadDiameterByLevel::CreateAndCompute( *p_arr, 0.0 ) );

    if   ( p_arr->size() < 4000 ) {
        printf( "Brute force...\n" );
        tester.push( BruteDiameter::CreateAndCompute( *p_arr ) );

        printf( "Brute force local...\n" );
        tester.push( BruteDiameterLocal::CreateAndCompute( *p_arr ) );
    }

    //-----------------------------------------------------------------
    // 0.01
    //-----------------------------------------------------------------
    printf( "Quad by heap (0.01)...\n" );
    tester.push( alg = QuadDiameterByHeap::CreateAndCompute( *p_arr, 0.01 ) );

    printf( "Quad by heap gdiam ptr ...\n" );
    tester.push( alg = QuadDiameterByGPtrHeap::CreateAndCompute( 
                                                   *p_arr, 0.01 ) );

    printf( "Quad by heap WSPD (0.01)...\n" );
    tester.push( alg = QuadDiameterByHeapWSPD::CreateAndCompute( *p_arr,
                                                                 0.01 ) );
    printf( "Quad by level_ptr 0.01...\n" );
    tester.push( QuadDiameterPtrByLevel::CreateAndCompute( *p_arr, 0.01 ) );

    printf( "Quad by level (0.01)...\n" );
    tester.push( QuadDiameterByLevel::CreateAndCompute( *p_arr, 0.01 ) );
    fflush( stdout );

    printf( "Chan algorithm (0.01)...\n" );
    tester.push( DiameterChan::CreateAndCompute( *p_arr, 0.01 ) );
    fflush( stdout );

    printf( "Chan modified algorithm (0.01)...\n" );
    tester.push( DiameterModChan::CreateAndCompute( *p_arr, 0.01 ) );
    fflush( stdout );

    /*
    if  ( p_arr->size() <= 3000 ) {
        printf( "Quad by directions (0.01)...\n" );
        tester.push( DiameterDirections::CreateAndCompute( *p_arr, 0.01 ) );
        fflush( stdout );
        }*/
    printf( "Quad by grid (0.01)...\n" );
    tester.push( DiameterGrid::CreateAndCompute( *p_arr, 0.01 ) );

    //printf( "Quad by grid + fs dir (0.01)...\n" );
    //tester.push( DiameterGridQuadDir::CreateAndCompute( *p_arr, 0.01 ) );

    printf( "Quad by dir (0.01)...\n" );
    tester.push( QuadDiameterByDir::CreateAndCompute( *p_arr, 0.01 ) );
    fflush( stdout );

    //-----------------------------------------------------------------
    // 0.1
    //-----------------------------------------------------------------
    printf( "Quad by heap (0.1)...\n" );
    tester.push( alg = QuadDiameterByHeap::CreateAndCompute( *p_arr, 0.1 ) );

    printf( "Quad by heap gdiam ptr ...\n" );
    tester.push( alg = QuadDiameterByGPtrHeap::CreateAndCompute( 
                                                   *p_arr, 0.1 ) );

    printf( "Quad by heap WSPD (0.1)...\n" );
    tester.push( alg = QuadDiameterByHeapWSPD::CreateAndCompute( *p_arr, 
                                                                 0.1 ) );

    printf( "Quad by level_ptr 0.1...\n" );
    tester.push( QuadDiameterPtrByLevel::CreateAndCompute( *p_arr, 0.1 ) );

    printf( "Quad by level (0.1)...\n" );
    tester.push( QuadDiameterByLevel::CreateAndCompute( *p_arr, 0.1 ) );
    printf( "Chan algorithm (0.1)...\n" );
    tester.push( DiameterChan::CreateAndCompute( *p_arr, 0.1 ) );
    fflush( stdout );

    printf( "Chan modified algorithm (0.1)...\n" );
    tester.push( DiameterModChan::CreateAndCompute( *p_arr, 0.1 ) );
    fflush( stdout );

    /*
    if  ( p_arr->size() <= 3  0 ) {
        printf( "Quad by directions (0.1)...\n" );
        tester.push( DiameterDirections::CreateAndCompute( *p_arr, 0.1 ) );
        fflush( stdout );
        }*/

    printf( "Quad by grid (0.1)...\n" );
    tester.push( DiameterGrid::CreateAndCompute( *p_arr, 0.1 ) );
    printf( "Quad by grid + fs dir (0.1)...\n" );
    tester.push( DiameterGridQuadDir::CreateAndCompute( *p_arr, 0.1 ) );
    printf( "Quad by dir (0.1)...\n" );
    fflush( stdout );
    tester.push( QuadDiameterByDir::CreateAndCompute( *p_arr, 0.1 ) );

    //-----------------------------------------------------------------
    // 1.72
    //-----------------------------------------------------------------
    //if  ( p_arr->size() <= 30000 ) {
        //printf( "PCA Mass algorithm ...\n" );
        //tester.push( DiameterMassPCA::CreateAndCompute( *p_arr ) );
        //fflush( stdout );
    //}
    printf( "Diameter by axis parallel bounding box\n" );
    fflush( stdout );
    tester.push( DiameterAPBBox::CreateAndCompute( *p_arr ) );

    printf( "PCA algorithm ...\n" );
    tester.push( DiameterPCA::CreateAndCompute( *p_arr ) );
    fflush( stdout );

    printf( "Points: %d\n", p_arr->size() );
    tester.dump_to_file( "out.txt", description );
    tester.dump();

    fflush( stdout );
    printf( "\n\nTest done!\n\n" );

    return   new  PointPair( alg->diam );
}

void  tester( const char  * file_name,
              const char  * description )
{
    ArrPoint3d   * p_arr;

    p_arr = new ArrPoint3d;
    p_arr->init( 10, 43283 );

    //p_arr->reset();

    if  ( ( suffix_eq( file_name, ".3ds" ) )
          ||  ( suffix_eq( file_name, ".3DS" ) ) )
        read_3ds( file_name, *p_arr );
    else
        read_poly_file( *p_arr, file_name );

    pnt_array_normalize( *p_arr );
    
    tester_inner( p_arr, description );

    delete p_arr;
}



void   compute_mvbb( const ArrPoint3d  & arr )
{
    Timer  tm;
    gdiam_real  * p_arr, * p_start;
    GBBox  ap_bbox;
    //double  num;

    tm.start();

    p_arr = (gdiam_real *)malloc( sizeof( gdiam_real ) * arr.size() * 3 );
    p_start = p_arr;
    for  ( int  ind = 0; ind < arr.size(); ind++ ) {
        p_arr[ 0 ] = arr[ ind ][ 0 ];
        p_arr[ 1 ] = arr[ ind ][ 1 ];
        p_arr[ 2 ] = arr[ ind ][ 2 ];
        p_arr += 3;
    }

    
    gdiam_bbox  bb, bb1;
    gdiam_point  * ptr_arr;

    Timer  tm_reg;

    tm_reg.start();

    ptr_arr = gdiam_convert( p_start, arr.size() );
    bb = gdiam_approx_const_mvbb( ptr_arr, arr.size(), 0, &ap_bbox );

    tm_reg.end();

    Timer  tm_reg2;

    tm_reg2.start();

    //bb1 = gdiam_approx_mvbb_grid( ptr_arr, arr.size(), 5 );
    bb1 = gdiam_approx_mvbb_grid_sample( ptr_arr, arr.size(), 
                                         5, 400 );
    
    tm_reg2.end();

    free( ptr_arr );
    free( p_start );
    
    tm.end();

    printf( "Overall time: %g\n", tm.seconds() );
    printf( "Time for const approx   : %g\n", tm_reg.seconds() );

    printf( "Volume on axis para bbox  : %g\n", ap_bbox.volume() );
            
    printf( "Volume of const approx    : %g\n", bb.volume() );
    printf( "Volume of grid            : %g\n", bb1.volume() );
    printf( "\n\n\n\n" );
    fflush( stdout );
}



/* algorithm.C - End of File ------------------------------------------*/








